from typing import Any, Dict, Optional

from netmiko import ConnectHandler

from nornir.core.configuration import Config
from nornir.core.connections import ConnectionPlugin

napalm_to_netmiko_map = {
    "ios": "cisco_ios",
    "nxos": "cisco_nxos",
    "nxos_ssh": "cisco_nxos",
    "eos": "arista_eos",
    "junos": "juniper_junos",
    "iosxr": "cisco_xr",
}


class Netmiko(ConnectionPlugin):
    """
    This plugin connects to the device using the Netmiko driver and sets the
    relevant connection.

    Inventory:
        extras: maps to argument passed to ``ConnectHandler``.
    """

    def open(
        self,
        hostname: Optional[str],
        username: Optional[str],
        password: Optional[str],
        port: Optional[int],
        platform: Optional[str],
        extras: Optional[Dict[str, Any]] = None,
        configuration: Optional[Config] = None,
    ) -> None:
        parameters = {
            "host": hostname,
            "username": username,
            "password": password,
            "port": port,
        }

        try:
            parameters[
                "ssh_config_file"
            ] = configuration.ssh.config_file  # type: ignore
        except AttributeError:
            pass

        if platform is not None:
            # Look platform up in corresponding map, if no entry return the host.nos unmodified
            platform = napalm_to_netmiko_map.get(platform, platform)
            parameters["device_type"] = platform

        extras = extras or {}
        parameters.update(extras)
        self.connection = ConnectHandler(**parameters)

    def close(self) -> None:
        self.connection.disconnect()



import shlex
import subprocess


from nornir.core.exceptions import CommandError
from nornir.core.task import Result, Task


def command(task: Task, command: str) -> Result:
    """
    Executes a command locally

    Arguments:
        command: command to execute

    Returns:
        Result object with the following attributes set:
          * result (``str``): stderr or stdout
          * stdout (``str``): stdout
          * stderr (``str``): stderr

    Raises:
        :obj:`nornir.core.exceptions.CommandError`: when there is a command error
    """
    cmd = subprocess.Popen(
        shlex.split(command),
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        shell=False,
    )

    stdout, stderr = cmd.communicate()
    stdout = stdout.decode()
    stderr = stderr.decode()

    if cmd.poll():
        raise CommandError(command, cmd.returncode, stdout, stderr)

    result = stderr if stderr else stdout
    return Result(result=result, host=task.host, stderr=stderr, stdout=stdout)



from typing import Any

from nornir.core.task import Result, Task


def netmiko_send_command(
    task: Task,
    command_string: str,
    use_timing: bool = False,
    enable: bool = False,
    **kwargs: Any
) -> Result:
    """
    Execute Netmiko send_command method (or send_command_timing)

    Arguments:
        command_string: Command to execute on the remote network device.
        use_timing: Set to True to switch to send_command_timing method.
        enable: Set to True to force Netmiko .enable() call.
        kwargs: Additional arguments to pass to send_command method.

    Returns:
        Result object with the following attributes set:
          * result: Result of the show command (generally a string, but depends on use of TextFSM).
    """
    net_connect = task.host.get_connection("netmiko", task.nornir.config)
    if enable:
        net_connect.enable()
    if use_timing:
        result = net_connect.send_command_timing(command_string, **kwargs)
    else:
        result = net_connect.send_command(command_string, **kwargs)
    return Result(host=task.host, result=result)



from typing import Any, Optional, Dict, Callable

from nornir.core.helpers import jinja_helper
from nornir.core.task import Result, Task

FiltersDict = Optional[Dict[str, Callable[..., str]]]


def template_file(
    task: Task,
    template: str,
    path: str,
    jinja_filters: FiltersDict = None,
    **kwargs: Any
) -> Result:
    """
    Renders contants of a file with jinja2. All the host data is available in the template

    Arguments:
        template: filename
        path: path to dir with templates
        jinja_filters: jinja filters to enable. Defaults to nornir.config.jinja2.filters
        **kwargs: additional data to pass to the template

    Returns:
        Result object with the following attributes set:
          * result (``string``): rendered string
    """
    jinja_filters = jinja_filters or {} or task.nornir.config.jinja2.filters
    text = jinja_helper.render_from_file(
        template=template,
        path=path,
        host=task.host,
        jinja_filters=jinja_filters,
        **kwargs
    )
    return Result(host=task.host, result=text)



import os
from typing import Any, Dict, Optional

from nornir.core.configuration import Config
from nornir.core.connections import ConnectionPlugin

import paramiko


class Paramiko(ConnectionPlugin):
    """
    This plugin connects to the device with paramiko to the device and sets the
    relevant connection.

    Inventory:
        extras: maps to argument passed to ``ConnectHandler``.
    """

    def open(
        self,
        hostname: Optional[str],
        username: Optional[str],
        password: Optional[str],
        port: Optional[int],
        platform: Optional[str],
        extras: Optional[Dict[str, Any]] = None,
        configuration: Optional[Config] = None,
    ) -> None:
        extras = extras or {}

        client = paramiko.SSHClient()
        client._policy = paramiko.WarningPolicy()
        client.set_missing_host_key_policy(paramiko.AutoAddPolicy())

        ssh_config = paramiko.SSHConfig()
        ssh_config_file = configuration.ssh.config_file  # type: ignore
        if os.path.exists(ssh_config_file):
            with open(ssh_config_file) as f:
                ssh_config.parse(f)
        parameters = {
            "hostname": hostname,
            "username": username,
            "password": password,
            "port": port,
        }

        user_config = ssh_config.lookup(hostname)
        for k in ("hostname", "username", "port"):
            if k in user_config:
                parameters[k] = user_config[k]

        if "proxycommand" in user_config:
            parameters["sock"] = paramiko.ProxyCommand(user_config["proxycommand"])

        self.state["ssh_forward_agent"] = user_config.get("forwardagent") == "yes"

        # TODO configurable
        #  if ssh_key_file:
        #      parameters['key_filename'] = ssh_key_file
        if "identityfile" in user_config:
            parameters["key_filename"] = user_config["identityfile"]

        extras.update(parameters)
        client.connect(**extras)
        self.connection = client

    def close(self) -> None:
        self.connection.close()



from nornir.core.task import Result, Task

import ruamel.yaml


def load_yaml(task: Task, file: str) -> Result:
    """
    Loads a yaml file.

    Arguments:
        file: path to the file containing the yaml file to load

    Examples:

        Simple example with ``ordered_dict``::

            > nr.run(task=load_yaml,
                     file="mydata.yaml")

    Returns:
        Result object with the following attributes set:
          * result (``dict``): dictionary with the contents of the file
    """
    with open(file, "r") as f:
        yml = ruamel.yaml.YAML(typ="safe")
        data = yml.load(f)

    return Result(host=task.host, result=data)



import logging
import os
from typing import Any, Optional

from nornir.core.deserializer.inventory import (
    HostsDict,
    GroupsDict,
    Inventory,
    VarsDict,
)

import ruamel.yaml

logger = logging.getLogger(__name__)


class SimpleInventory(Inventory):
    def __init__(
        self,
        host_file: str = "hosts.yaml",
        group_file: str = "groups.yaml",
        defaults_file: str = "defaults.yaml",
        hosts: Optional[HostsDict] = None,
        groups: Optional[GroupsDict] = None,
        defaults: Optional[VarsDict] = None,
        *args: Any,
        **kwargs: Any
    ) -> None:
        if hosts is None:
            yml = ruamel.yaml.YAML(typ="safe")
            with open(host_file, "r") as f:
                hosts = yml.load(f)

        if groups is None:
            groups = {}
            if group_file:
                if os.path.exists(group_file):
                    with open(group_file, "r") as f:
                        groups = yml.load(f) or {}
                else:
                    logger.debug("File %r was not found", group_file)
                    groups = {}

        if defaults is None:
            defaults = {}
            if defaults_file:
                if os.path.exists(defaults_file):
                    with open(defaults_file, "r") as f:
                        defaults = yml.load(f) or {}
                else:
                    logger.debug("File %r was not found", defaults_file)
                    defaults = {}
        super().__init__(hosts=hosts, groups=groups, defaults=defaults, *args, **kwargs)



from abc import ABC, abstractmethod
from typing import Any, Dict, NoReturn, Optional, Type


from nornir.core.configuration import Config
from nornir.core.exceptions import (
    ConnectionPluginAlreadyRegistered,
    ConnectionPluginNotRegistered,
)


class ConnectionPlugin(ABC):
    """
    Connection plugins have to inherit from this class and provide implementations
    for both the :meth:`open` and :meth:`close` methods.

    Attributes:
        connection: Underlying connection. Populated by :meth:`open`.
        state: Dictionary to hold any data that needs to be shared between
            the connection plugin and the plugin tasks using this connection.
    """

    __slots__ = ("connection", "state")

    def __init__(self) -> None:
        self.connection: Any = UnestablishedConnection()
        self.state: Dict[str, Any] = {}

    @abstractmethod
    def open(
        self,
        hostname: Optional[str],
        username: Optional[str],
        password: Optional[str],
        port: Optional[int],
        platform: Optional[str],
        extras: Optional[Dict[str, Any]] = None,
        configuration: Optional[Config] = None,
    ) -> None:
        """
        Connect to the device and populate the attribute :attr:`connection` with
        the underlying connection
        """
        pass

    @abstractmethod
    def close(self) -> None:
        """Close the connection with the device"""
        pass


class UnestablishedConnection(object):
    def close(self) -> NoReturn:
        raise ValueError("Connection not established")

    disconnect = close


class Connections(Dict[str, ConnectionPlugin]):
    available: Dict[str, Type[ConnectionPlugin]] = {}

    @classmethod
    def register(cls, name: str, plugin: Type[ConnectionPlugin]) -> None:
        """Registers a connection plugin with a specified name

        Args:
            name: name of the connection plugin to register
            plugin: defined connection plugin class

        Raises:
            :obj:`nornir.core.exceptions.ConnectionPluginAlreadyRegistered` if
                another plugin with the specified name was already registered
        """
        existing_plugin = cls.available.get(name)
        if existing_plugin is None:
            cls.available[name] = plugin
        elif existing_plugin != plugin:
            raise ConnectionPluginAlreadyRegistered(
                f"Connection plugin {plugin.__name__} can't be registered as "
                f"{name!r} because plugin {existing_plugin.__name__} "
                f"was already registered under this name"
            )

    @classmethod
    def deregister(cls, name: str) -> None:
        """Deregisters a registered connection plugin by its name

        Args:
            name: name of the connection plugin to deregister

        Raises:
            :obj:`nornir.core.exceptions.ConnectionPluginNotRegistered`
        """
        if name not in cls.available:
            raise ConnectionPluginNotRegistered(
                f"Connection {name!r} is not registered"
            )
        cls.available.pop(name)

    @classmethod
    def deregister_all(cls) -> None:
        """Deregisters all registered connection plugins"""
        cls.available = {}

    @classmethod
    def get_plugin(cls, name: str) -> Type[ConnectionPlugin]:
        """Fetches the connection plugin by name if already registered

        Args:
            name: name of the connection plugin

        Raises:
            :obj:`nornir.core.exceptions.ConnectionPluginNotRegistered`
        """
        if name not in cls.available:
            raise ConnectionPluginNotRegistered(
                f"Connection {name!r} is not registered"
            )
        return cls.available[name]



import logging
import pprint
import threading
from typing import List, Optional, cast
from collections import OrderedDict
import json

from colorama import Fore, Style, init

from nornir.core.task import AggregatedResult, MultiResult, Result


LOCK = threading.Lock()


init(autoreset=True, strip=False)


def print_title(title: str) -> None:
    """
    Helper function to print a title.
    """
    msg = "**** {} ".format(title)
    print("{}{}{}{}".format(Style.BRIGHT, Fore.GREEN, msg, "*" * (80 - len(msg))))


def _get_color(result: Result, failed: bool) -> str:
    if result.failed or failed:
        color = Fore.RED
    elif result.changed:
        color = Fore.YELLOW
    else:
        color = Fore.GREEN
    return cast(str, color)


def _print_individual_result(
    result: Result,
    host: Optional[str],
    attrs: List[str],
    failed: bool,
    severity_level: int,
    task_group: bool = False,
) -> None:
    if result.severity_level < severity_level:
        return

    color = _get_color(result, failed)
    subtitle = (
        "" if result.changed is None else " ** changed : {} ".format(result.changed)
    )
    level_name = logging.getLevelName(result.severity_level)
    symbol = "v" if task_group else "-"
    msg = "{} {}{}".format(symbol * 4, result.name, subtitle)
    print(
        "{}{}{}{} {}".format(
            Style.BRIGHT, color, msg, symbol * (80 - len(msg)), level_name
        )
    )
    for attribute in attrs:
        x = getattr(result, attribute, "")
        if isinstance(x, BaseException):
            # for consistency between py3.6 and py3.7
            print(f"{x.__class__.__name__}{x.args}")
        elif x and not isinstance(x, str):
            if isinstance(x, OrderedDict):
                print(json.dumps(x, indent=2))
            else:
                pprint.pprint(x, indent=2)
        elif x:
            print(x)


def _print_result(
    result: Result,
    host: Optional[str] = None,
    attrs: List[str] = None,
    failed: bool = False,
    severity_level: int = logging.INFO,
) -> None:
    attrs = attrs or ["diff", "result", "stdout"]
    if isinstance(attrs, str):
        attrs = [attrs]

    if isinstance(result, AggregatedResult):
        msg = result.name
        print("{}{}{}{}".format(Style.BRIGHT, Fore.CYAN, msg, "*" * (80 - len(msg))))
        for host, host_data in sorted(result.items()):
            title = (
                ""
                if host_data.changed is None
                else " ** changed : {} ".format(host_data.changed)
            )
            msg = "* {}{}".format(host, title)
            print(
                "{}{}{}{}".format(Style.BRIGHT, Fore.BLUE, msg, "*" * (80 - len(msg)))
            )
            _print_result(host_data, host, attrs, failed, severity_level)
    elif isinstance(result, MultiResult):
        _print_individual_result(
            result[0], host, attrs, failed, severity_level, task_group=True
        )
        for r in result[1:]:
            _print_result(r, host, attrs, failed, severity_level)
        color = _get_color(result[0], failed)
        msg = "^^^^ END {} ".format(result[0].name)
        print("{}{}{}{}".format(Style.BRIGHT, color, msg, "^" * (80 - len(msg))))
    elif isinstance(result, Result):
        _print_individual_result(result, host, attrs, failed, severity_level)


def print_result(
    result: Result,
    host: Optional[str] = None,
    vars: List[str] = None,
    failed: bool = False,
    severity_level: int = logging.INFO,
) -> None:
    """
    Prints the :obj:`nornir.core.task.Result` from a previous task to screen

    Arguments:
        result: from a previous task
        host: # TODO
        vars: Which attributes you want to print
        failed: if ``True`` assume the task failed
        severity_level: Print only errors with this severity level or higher
    """
    LOCK.acquire()
    try:
        _print_result(result, host, vars, failed, severity_level)
    finally:
        LOCK.release()



from typing import Any

from netmiko import file_transfer

from nornir.core.task import Result, Task


def netmiko_file_transfer(
    task: Task, source_file: str, dest_file: str, **kwargs: Any
) -> Result:
    """
    Execute Netmiko file_transfer method

    Arguments:
        source_file: Source file.
        dest_file: Destination file.
        kwargs: Additional arguments to pass to file_transfer

    Returns:
        Result object with the following attributes set:
          * result (``bool``): file exists and MD5 is valid
          * changed (``bool``): the destination file was changed

    """
    net_connect = task.host.get_connection("netmiko", task.nornir.config)
    kwargs.setdefault("direction", "put")
    scp_result = file_transfer(
        net_connect, source_file=source_file, dest_file=dest_file, **kwargs
    )
    if kwargs.get("disable_md5") is True:
        file_valid = scp_result["file_exists"]
    else:
        file_valid = scp_result["file_exists"] and scp_result["file_verified"]
    return Result(
        host=task.host, result=file_valid, changed=scp_result["file_transferred"]
    )



import json
from typing import Any, Dict, MutableMapping, Type

from nornir.core.task import Result, Task


def load_json(task: Task, file: str) -> Result:
    """
    Loads a json file.

    Arguments:
        file: path to the file containing the json file to load

    Examples:

        Simple example with ``ordered_dict``::

            > nr.run(task=load_json,
                     file="mydata.json")

        file: path to the file containing the json file to load

    Returns:
        Result object with the following attributes set:
          * result (``dict``): dictionary with the contents of the file
    """
    kwargs: Dict[str, Type[MutableMapping[str, Any]]] = {}
    with open(file, "r") as f:
        data = json.loads(f.read(), **kwargs)

    return Result(host=task.host, result=data)



import warnings
from collections import UserList
from typing import Any, Dict, List, Optional, Set, Union

from nornir.core import deserializer
from nornir.core.configuration import Config
from nornir.core.connections import ConnectionPlugin, Connections
from nornir.core.exceptions import ConnectionAlreadyOpen, ConnectionNotOpen


class BaseAttributes(object):
    __slots__ = ("hostname", "port", "username", "password", "platform")

    def __init__(
        self,
        hostname: Optional[str] = None,
        port: Optional[int] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        platform: Optional[str] = None,
    ) -> None:
        self.hostname = hostname
        self.port = port
        self.username = username
        self.password = password
        self.platform = platform

    def dict(self):
        w = f"{self.dict.__qualname__} is deprecated, use nornir.core.deserializer instead"
        warnings.warn(w)
        return (
            getattr(deserializer.inventory, self.__class__.__name__)
            .serialize(self)
            .dict()
        )


class ConnectionOptions(BaseAttributes):
    __slots__ = ("extras",)

    def __init__(self, extras: Optional[Dict[str, Any]] = None, **kwargs) -> None:
        self.extras = extras
        super().__init__(**kwargs)


class ParentGroups(UserList):
    __slots__ = "refs"

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.refs: List["Group"] = kwargs.get("refs", [])

    def __contains__(self, value) -> bool:
        return value in self.data or value in self.refs


class InventoryElement(BaseAttributes):
    __slots__ = ("groups", "data", "connection_options")

    def __init__(
        self,
        groups: Optional[ParentGroups] = None,
        data: Optional[Dict[str, Any]] = None,
        connection_options: Optional[Dict[str, ConnectionOptions]] = None,
        **kwargs,
    ) -> None:
        self.groups = groups or ParentGroups()
        self.data = data or {}
        self.connection_options = connection_options or {}
        super().__init__(**kwargs)


class Defaults(BaseAttributes):
    __slots__ = ("data", "connection_options")

    def __init__(
        self,
        data: Optional[Dict[str, Any]] = None,
        connection_options: Optional[Dict[str, ConnectionOptions]] = None,
        **kwargs,
    ) -> None:
        self.data = data or {}
        self.connection_options = connection_options or {}
        super().__init__(**kwargs)


class Host(InventoryElement):
    __slots__ = ("name", "connections", "defaults")

    def __init__(
        self, name: str, defaults: Optional[Defaults] = None, **kwargs
    ) -> None:
        self.name = name
        self.defaults = defaults or Defaults()
        self.connections: Connections = Connections()
        super().__init__(**kwargs)

    def _resolve_data(self):
        processed = []
        result = {}
        for k, v in self.data.items():
            processed.append(k)
            result[k] = v
        for g in self.groups.refs:
            for k, v in g.items():
                if k not in processed:
                    processed.append(k)
                    result[k] = v
        for k, v in self.defaults.data.items():
            if k not in processed:
                processed.append(k)
                result[k] = v
        return result

    def keys(self):
        """Returns the keys of the attribute ``data`` and of the parent(s) groups."""
        return self._resolve_data().keys()

    def values(self):
        """Returns the values of the attribute ``data`` and of the parent(s) groups."""
        return self._resolve_data().values()

    def items(self):
        """
        Returns all the data accessible from a device, including
        the one inherited from parent groups
        """
        return self._resolve_data().items()

    def has_parent_group(self, group):
        """Retuns whether the object is a child of the :obj:`Group` ``group``"""
        if isinstance(group, str):
            return self._has_parent_group_by_name(group)

        else:
            return self._has_parent_group_by_object(group)

    def _has_parent_group_by_name(self, group):
        for g in self.groups.refs:
            if g.name == group or g.has_parent_group(group):
                return True

    def _has_parent_group_by_object(self, group):
        for g in self.groups.refs:
            if g is group or g.has_parent_group(group):
                return True

    def __getitem__(self, item):
        try:
            return self.data[item]

        except KeyError:
            for g in self.groups.refs:
                try:
                    r = g[item]
                    return r
                except KeyError:
                    continue

            r = self.defaults.data.get(item)
            if r:
                return r

            raise

    def __getattribute__(self, name):
        if name not in ("hostname", "port", "username", "password", "platform"):
            return object.__getattribute__(self, name)
        v = object.__getattribute__(self, name)
        if v is None:
            for g in self.groups.refs:
                r = getattr(g, name)
                if r is not None:
                    return r

            return object.__getattribute__(self.defaults, name)
        else:
            return v

    def __bool__(self):
        return bool(self.name)

    def __setitem__(self, item, value):
        self.data[item] = value

    def __len__(self):
        return len(self._resolve_data().keys())

    def __iter__(self):
        return self.data.__iter__()

    def __str__(self):
        return self.name

    def __repr__(self):
        return "{}: {}".format(self.__class__.__name__, self.name or "")

    def get(self, item, default=None):
        """
        Returns the value ``item`` from the host or hosts group variables.

        Arguments:
            item(``str``): The variable to get
            default(``any``): Return value if item not found
        """
        if hasattr(self, item):
            return getattr(self, item)
        try:
            return self.__getitem__(item)

        except KeyError:
            return default

    def get_connection_parameters(
        self, connection: Optional[str] = None
    ) -> ConnectionOptions:
        if not connection:
            d = ConnectionOptions(
                hostname=self.hostname,
                port=self.port,
                username=self.username,
                password=self.password,
                platform=self.platform,
                extras={},
            )
        else:
            r = self._get_connection_options_recursively(connection)
            if r is not None:
                d = ConnectionOptions(
                    hostname=r.hostname if r.hostname is not None else self.hostname,
                    port=r.port if r.port is not None else self.port,
                    username=r.username if r.username is not None else self.username,
                    password=r.password if r.password is not None else self.password,
                    platform=r.platform if r.platform is not None else self.platform,
                    extras=r.extras if r.extras is not None else {},
                )
            else:
                d = ConnectionOptions(
                    hostname=self.hostname,
                    port=self.port,
                    username=self.username,
                    password=self.password,
                    platform=self.platform,
                    extras={},
                )
        return d

    def _get_connection_options_recursively(
        self, connection: str
    ) -> Optional[ConnectionOptions]:
        p = self.connection_options.get(connection)
        if p is None:
            p = ConnectionOptions()

        for g in self.groups.refs:
            sp = g._get_connection_options_recursively(connection)
            if sp is not None:
                p.hostname = p.hostname if p.hostname is not None else sp.hostname
                p.port = p.port if p.port is not None else sp.port
                p.username = p.username if p.username is not None else sp.username
                p.password = p.password if p.password is not None else sp.password
                p.platform = p.platform if p.platform is not None else sp.platform
                p.extras = p.extras if p.extras is not None else sp.extras

        sp = self.defaults.connection_options.get(connection, None)
        if sp is not None:
            p.hostname = p.hostname if p.hostname is not None else sp.hostname
            p.port = p.port if p.port is not None else sp.port
            p.username = p.username if p.username is not None else sp.username
            p.password = p.password if p.password is not None else sp.password
            p.platform = p.platform if p.platform is not None else sp.platform
            p.extras = p.extras if p.extras is not None else sp.extras
        return p

    def get_connection(self, connection: str, configuration: Config) -> Any:
        """
        The function of this method is twofold:

            1. If an existing connection is already established for the given type return it
            2. If none exists, establish a new connection of that type with default parameters
               and return it

        Raises:
            AttributeError: if it's unknown how to establish a connection for the given type

        Arguments:
            connection: Name of the connection, for instance, netmiko, paramiko, napalm...

        Returns:
            An already established connection
        """
        if connection not in self.connections:
            conn = self.get_connection_parameters(connection)
            self.open_connection(
                connection=connection,
                configuration=configuration,
                hostname=conn.hostname,
                port=conn.port,
                username=conn.username,
                password=conn.password,
                platform=conn.platform,
                extras=conn.extras,
            )
        return self.connections[connection].connection

    def get_connection_state(self, connection: str) -> Dict[str, Any]:
        """
        For an already established connection return its state.
        """
        if connection not in self.connections:
            raise ConnectionNotOpen(connection)

        return self.connections[connection].state

    def open_connection(
        self,
        connection: str,
        configuration: Config,
        hostname: Optional[str] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        port: Optional[int] = None,
        platform: Optional[str] = None,
        extras: Optional[Dict[str, Any]] = None,
        default_to_host_attributes: bool = True,
    ) -> ConnectionPlugin:
        """
        Open a new connection.

        If ``default_to_host_attributes`` is set to ``True`` arguments will default to host
        attributes if not specified.

        Raises:
            AttributeError: if it's unknown how to establish a connection for the given type

        Returns:
            An already established connection
        """
        if connection in self.connections:
            raise ConnectionAlreadyOpen(connection)

        self.connections[connection] = self.connections.get_plugin(connection)()
        if default_to_host_attributes:
            conn_params = self.get_connection_parameters(connection)
            self.connections[connection].open(
                hostname=hostname if hostname is not None else conn_params.hostname,
                username=username if username is not None else conn_params.username,
                password=password if password is not None else conn_params.password,
                port=port if port is not None else conn_params.port,
                platform=platform if platform is not None else conn_params.platform,
                extras=extras if extras is not None else conn_params.extras,
                configuration=configuration,
            )
        else:
            self.connections[connection].open(
                hostname=hostname,
                username=username,
                password=password,
                port=port,
                platform=platform,
                extras=extras,
                configuration=configuration,
            )
        return self.connections[connection]

    def close_connection(self, connection: str) -> None:
        """ Close the connection"""
        if connection not in self.connections:
            raise ConnectionNotOpen(connection)

        self.connections.pop(connection).close()

    def close_connections(self) -> None:
        # Decouple deleting dictionary elements from iterating over connections dict
        existing_conns = list(self.connections.keys())
        for connection in existing_conns:
            self.close_connection(connection)


class Group(Host):
    pass


class Hosts(Dict[str, Host]):
    pass


class Groups(Dict[str, Group]):
    pass


class Inventory(object):
    __slots__ = ("hosts", "groups", "defaults")

    def __init__(
        self,
        hosts: Hosts,
        groups: Optional[Groups] = None,
        defaults: Optional[Defaults] = None,
        transform_function=None,
        transform_function_options=None,
    ) -> None:
        self.hosts = hosts
        self.groups = groups or Groups()
        self.defaults = defaults or Defaults()

        for host in self.hosts.values():
            host.groups.refs = [self.groups[p] for p in host.groups]
        for group in self.groups.values():
            group.groups.refs = [self.groups[p] for p in group.groups]

        if transform_function:
            for h in self.hosts.values():
                transform_function(h, **transform_function_options)

    def filter(self, filter_obj=None, filter_func=None, *args, **kwargs):
        filter_func = filter_obj or filter_func
        if filter_func:
            filtered = {n: h for n, h in self.hosts.items() if filter_func(h, **kwargs)}
        else:
            filtered = {
                n: h
                for n, h in self.hosts.items()
                if all(h.get(k) == v for k, v in kwargs.items())
            }
        return Inventory(hosts=filtered, groups=self.groups, defaults=self.defaults)

    def __len__(self):
        return self.hosts.__len__()

    def children_of_group(self, group: Union[str, Group]) -> Set[Host]:
        """
        Returns set of hosts that belongs to a group including those that belong
        indirectly via inheritance
        """
        hosts: List[Host] = set()
        for host in self.hosts.values():
            if host.has_parent_group(group):
                hosts.add(host)
        return hosts

    def add_host(self, name: str, **kwargs) -> None:
        """
        Add a host to the inventory after initialization
        """
        host = {
            name: deserializer.inventory.InventoryElement.deserialize_host(
                name=name, defaults=self.defaults, **kwargs
            )
        }
        self.hosts.update(host)

    def add_group(self, name: str, **kwargs) -> None:
        """
        Add a group to the inventory after initialization
        """
        group = {
            name: deserializer.inventory.InventoryElement.deserialize_group(
                name=name, defaults=self.defaults, **kwargs
            )
        }
        self.groups.update(group)

    def get_inventory_dict(self) -> Dict:
        """
        Return serialized dictionary of inventory
        """
        return deserializer.inventory.Inventory.serialize(self).dict()

    def get_defaults_dict(self) -> Dict:
        """
        Returns serialized dictionary of defaults from inventory
        """
        return deserializer.inventory.Defaults.serialize(self.defaults).dict()

    def get_groups_dict(self) -> Dict:
        """
        Returns serialized dictionary of groups from inventory
        """
        return {
            k: deserializer.inventory.InventoryElement.serialize(v).dict()
            for k, v in self.groups.items()
        }

    def get_hosts_dict(self) -> Dict:
        """
        Returns serialized dictionary of hosts from inventory
        """
        return {
            k: deserializer.inventory.InventoryElement.serialize(v).dict()
            for k, v in self.hosts.items()
        }



import socket
from typing import Optional, List

from nornir.core.task import Result, Task


def tcp_ping(
    task: Task, ports: List[int], timeout: int = 2, host: Optional[str] = None
) -> Result:
    """
    Tests connection to a tcp port and tries to establish a three way
    handshake. To be used for network discovery or testing.

    Arguments:
        ports (list of int): tcp ports to ping
        timeout (int, optional): defaults to 2
        host (string, optional): defaults to ``hostname``


    Returns:
        Result object with the following attributes set:
          * result (``dict``): Contains port numbers as keys with True/False as values
    """

    if isinstance(ports, int):
        ports = [ports]

    if isinstance(ports, list):
        if not all(isinstance(port, int) for port in ports):
            raise ValueError("Invalid value for 'ports'")

    else:
        raise ValueError("Invalid value for 'ports'")

    host = host or task.host.hostname

    result = {}
    for port in ports:
        s = socket.socket()
        s.settimeout(timeout)
        try:
            status = s.connect_ex((host, port))
            if status == 0:
                connection = True
            else:
                connection = False
        except (socket.gaierror, socket.timeout, socket.error):
            connection = False
        finally:
            s.close()
        result[port] = connection

    return Result(host=task.host, result=result)



import configparser as cp
import logging
import os
from collections import defaultdict
from pathlib import Path
from typing import Any, DefaultDict, Dict, MutableMapping, Optional, Tuple, Union, cast

from mypy_extensions import TypedDict

from nornir.core.deserializer.inventory import (
    DefaultsDict,
    GroupsDict,
    HostsDict,
    Inventory,
    InventoryElement,
    VarsDict,
)

import ruamel.yaml
from ruamel.yaml.composer import ComposerError
from ruamel.yaml.scanner import ScannerError


VARS_FILENAME_EXTENSIONS = ["", ".yml", ".yaml"]


YAML = ruamel.yaml.YAML(typ="safe")

logger = logging.getLogger(__name__)


AnsibleHostsDict = Dict[str, Optional[VarsDict]]

AnsibleGroupDataDict = TypedDict(
    "AnsibleGroupDataDict",
    {"children": Dict[str, Any], "vars": VarsDict, "hosts": AnsibleHostsDict},
    total=False,
)  # bug: https://github.com/python/mypy/issues/5357

AnsibleGroupsDict = Dict[str, AnsibleGroupDataDict]


class AnsibleParser(object):
    def __init__(self, hostsfile: str) -> None:
        self.hostsfile = hostsfile
        self.path = os.path.dirname(hostsfile)
        self.hosts: HostsDict = {}
        self.groups: GroupsDict = {}
        self.defaults: DefaultsDict = {"data": {}}
        self.original_data: Optional[AnsibleGroupsDict] = None
        self.load_hosts_file()

    def parse_group(
        self, group: str, data: AnsibleGroupDataDict, parent: Optional[str] = None
    ) -> None:
        data = data or {}
        if group == "defaults":
            group_file = "all"
            dest_group = self.defaults
        else:
            self.add(group, self.groups)
            group_file = group
            dest_group = self.groups[group]

        if parent and parent != "defaults":
            dest_group["groups"].append(parent)

        group_data = data.get("vars", {})
        vars_file_data = self.read_vars_file(group_file, self.path, False) or {}
        self.normalize_data(dest_group, group_data, vars_file_data)
        self.map_nornir_vars(dest_group)

        self.parse_hosts(data.get("hosts", {}), parent=group)

        for children, children_data in data.get("children", {}).items():
            self.parse_group(
                children, cast(AnsibleGroupDataDict, children_data), parent=group
            )

    def parse(self) -> None:
        if self.original_data is not None:
            self.parse_group("defaults", self.original_data["all"])
        self.sort_groups()

    def parse_hosts(
        self, hosts: AnsibleHostsDict, parent: Optional[str] = None
    ) -> None:
        for host, data in hosts.items():
            data = data or {}
            self.add(host, self.hosts)
            if parent and parent != "defaults":
                self.hosts[host]["groups"].append(parent)

            vars_file_data = self.read_vars_file(host, self.path, True)
            self.normalize_data(self.hosts[host], data, vars_file_data)
            self.map_nornir_vars(self.hosts[host])

    def normalize_data(
        self, host: HostsDict, data: Dict[str, Any], vars_data: Dict[str, Any]
    ) -> None:
        reserved_fields = InventoryElement.__fields__.keys()
        self.map_nornir_vars(data)
        for k, v in data.items():
            if k in reserved_fields:
                host[k] = v
            else:
                host["data"][k] = v
        self.map_nornir_vars(vars_data)
        for k, v in vars_data.items():
            if k in reserved_fields:
                host[k] = v
            else:
                host["data"][k] = v

    def sort_groups(self) -> None:
        for host in self.hosts.values():
            host["groups"].sort()

        for name, group in self.groups.items():
            if name == "defaults":
                continue

            group["groups"].sort()

    @staticmethod
    def read_vars_file(element: str, path: str, is_host: bool = True) -> VarsDict:
        sub_dir = "host_vars" if is_host else "group_vars"
        vars_dir = Path(path) / sub_dir
        if vars_dir.is_dir():
            vars_file_base = vars_dir / element
            for extension in VARS_FILENAME_EXTENSIONS:
                vars_file = vars_file_base.with_suffix(
                    vars_file_base.suffix + extension
                )
                if vars_file.is_file():
                    with open(vars_file) as f:
                        logger.debug("AnsibleInventory: reading var file %r", vars_file)
                        return cast(Dict[str, Any], YAML.load(f))
            logger.debug(
                "AnsibleInventory: no vars file was found with the path %r "
                "and one of the supported extensions: %s",
                vars_file_base,
                VARS_FILENAME_EXTENSIONS,
            )
        return {}

    @staticmethod
    def map_nornir_vars(obj: VarsDict) -> None:
        mappings = {
            "ansible_host": "hostname",
            "ansible_port": "port",
            "ansible_user": "username",
            "ansible_password": "password",
        }
        for ansible_var, nornir_var in mappings.items():
            if ansible_var in obj:
                obj[nornir_var] = obj.pop(ansible_var)

    @staticmethod
    def add(element: str, element_dict: Dict[str, VarsDict]) -> None:
        if element not in element_dict:
            element_dict[element] = {"groups": [], "data": {}}

    def load_hosts_file(self) -> None:
        raise NotImplementedError


class INIParser(AnsibleParser):
    @staticmethod
    def normalize_value(value: str) -> Union[str, int]:
        try:
            return int(value)

        except (ValueError, TypeError):
            return value

    @staticmethod
    def normalize_content(content: str) -> VarsDict:
        result: VarsDict = {}

        if not content:
            return result

        for option in content.split():
            key, value = option.split("=")
            result[key] = INIParser.normalize_value(value)
        return result

    @staticmethod
    def process_meta(
        meta: Optional[str], section: MutableMapping[str, str]
    ) -> Dict[str, Any]:
        if meta == "vars":
            return {
                key: INIParser.normalize_value(value) for key, value in section.items()
            }

        elif meta == "children":
            return {group_name: {} for group_name in section}

        else:
            raise ValueError(f"Unknown tag {meta}")

    def normalize(self, data: cp.ConfigParser) -> Dict[str, AnsibleGroupDataDict]:
        groups: DefaultDict[str, Dict[str, Any]] = defaultdict(dict)
        # Dict[str, AnsibleGroupDataDict] does not work because of
        # https://github.com/python/mypy/issues/5359
        result: Dict[str, Dict[str, Dict[str, Dict[str, Any]]]] = {
            "all": {"children": groups}
        }

        for section_name, section in data.items():

            if section_name == "DEFAULT":
                continue

            if ":" in section_name:
                group_name, meta = section_name.split(":")
                subsection = self.process_meta(meta, section)
                if group_name == "all":
                    result["all"][meta] = subsection
                else:
                    groups[group_name][meta] = subsection
            else:
                groups[section_name]["hosts"] = {
                    host: self.normalize_content(host_vars)
                    for host, host_vars in section.items()
                }
        return cast(AnsibleGroupsDict, result)

    def load_hosts_file(self) -> None:
        original_data = cp.ConfigParser(
            interpolation=None, allow_no_value=True, delimiters=" ="
        )
        original_data.read(self.hostsfile)
        self.original_data = self.normalize(original_data)


class YAMLParser(AnsibleParser):
    def load_hosts_file(self) -> None:
        with open(self.hostsfile, "r") as f:
            self.original_data = cast(AnsibleGroupsDict, YAML.load(f))


def parse(hostsfile: str) -> Tuple[HostsDict, GroupsDict, DefaultsDict]:
    try:
        parser: AnsibleParser = INIParser(hostsfile)
    except cp.Error:
        try:
            parser = YAMLParser(hostsfile)
        except (ScannerError, ComposerError):
            logger.error("AnsibleInventory: file %r is not INI or YAML file", hostsfile)
            raise

    parser.parse()

    return parser.hosts, parser.groups, parser.defaults


class AnsibleInventory(Inventory):
    def __init__(self, hostsfile: str = "hosts", *args: Any, **kwargs: Any) -> None:
        host_vars, group_vars, defaults = parse(hostsfile)
        super().__init__(
            hosts=host_vars, groups=group_vars, defaults=defaults, *args, **kwargs
        )



from typing import Any

from nornir.core.task import Result, Task


def echo_data(task: Task, **kwargs: Any) -> Result:
    """
    Dummy task that echoes the data passed to it. Useful in grouped_tasks
    to debug data passed to tasks.

    Arguments:
        ``**kwargs``: Any <key,value> pair you want

    Returns:
        Result object with the following attributes set:
          * result (``dict``): ``**kwargs`` passed to the task
    """
    return Result(host=task.host, result=kwargs)



import warnings
from typing import Any, Callable, Dict, Optional

from nornir.core import Nornir
from nornir.core.connections import Connections
from nornir.core.deserializer.configuration import Config
from nornir.core.state import GlobalState
from nornir.plugins.connections.napalm import Napalm
from nornir.plugins.connections.netmiko import Netmiko
from nornir.plugins.connections.paramiko import Paramiko


def register_default_connection_plugins() -> None:
    Connections.register("napalm", Napalm)
    Connections.register("netmiko", Netmiko)
    Connections.register("paramiko", Paramiko)


def cls_to_string(cls: Callable[..., Any]) -> str:
    return f"{cls.__module__}.{cls.__name__}"


def InitNornir(
    config_file: str = "",
    dry_run: bool = False,
    configure_logging: Optional[bool] = None,
    **kwargs: Dict[str, Any],
) -> Nornir:
    """
    Arguments:
        config_file(str): Path to the configuration file (optional)
        dry_run(bool): Whether to simulate changes or not
        configure_logging: Whether to configure logging or not. This argument is being
            deprecated. Please use logging.enabled parameter in the configuration
            instead.
        **kwargs: Extra information to pass to the
            :obj:`nornir.core.configuration.Config` object

    Returns:
        :obj:`nornir.core.Nornir`: fully instantiated and configured
    """
    register_default_connection_plugins()

    if callable(kwargs.get("inventory", {}).get("plugin", "")):
        kwargs["inventory"]["plugin"] = cls_to_string(kwargs["inventory"]["plugin"])

    if callable(kwargs.get("inventory", {}).get("transform_function", "")):
        kwargs["inventory"]["transform_function"] = cls_to_string(
            kwargs["inventory"]["transform_function"]
        )

    conf = Config.load_from_file(config_file, **kwargs)

    data = GlobalState(dry_run=dry_run)

    if configure_logging is not None:
        msg = (
            "'configure_logging' argument is deprecated, please use "
            "'logging.enabled' parameter in the configuration instead: "
            "https://nornir.readthedocs.io/en/stable/configuration/index.html"
        )
        warnings.warn(msg, DeprecationWarning)

    if conf.logging.enabled is None:
        if configure_logging is not None:
            conf.logging.enabled = configure_logging
        else:
            conf.logging.enabled = True

    conf.logging.configure()

    inv = conf.inventory.plugin.deserialize(
        transform_function=conf.inventory.transform_function,
        transform_function_options=conf.inventory.transform_function_options,
        config=conf,
        **conf.inventory.options,
    )

    return Nornir(inventory=inv, config=conf, data=data)



from typing import Optional, Any
from nornir.core.task import Result, Task

import requests


def http_method(
    task: Optional[Task] = None,
    method: str = "get",
    url: str = "",
    raise_for_status: bool = True,
    **kwargs: Any
) -> Result:
    """
    This is a convenience task that uses `requests <http://docs.python-requests.org/en/master/>`_ to
    interact with an HTTP server.

    Arguments:
        method: HTTP method to call
        url: URL to connect to
        raise_for_status: Whether to call `raise_for_status
            <http://docs.python-requests.org/en/master/api/#requests.Response.raise_for_status>`_
            method automatically or not. For quick reference, raise_for_status will consider an
            error if the return code is any of 4xx or 5xx
        **kwargs: Keyword arguments will be passed to the `request
            <http://docs.python-requests.org/en/master/api/#requests.request>`_
            method

    Returns:
        Result object with the following attributes set:
          * result (``string/dict``): Body of the response. Either text or a dict if the
            response was a json object
          * response (``requests.Response``): Original `Response
            <http://docs.python-requests.org/en/master/api/#requests.Response>`_
    """
    r = requests.request(method, url, **kwargs)

    if raise_for_status:
        r.raise_for_status()

    try:
        content_type = r.headers["Content-type"]
    except KeyError:
        content_type = "text"

    result = r.json() if "application/json" == content_type else r.text

    return Result(host=task.host if task else None, response=r, result=result)



from typing import Any, Callable, Dict, List, Optional, Union

from nornir.core import inventory

from pydantic import BaseModel


VarsDict = Dict[str, Any]
HostsDict = Dict[str, VarsDict]
GroupsDict = Dict[str, VarsDict]
DefaultsDict = VarsDict


class BaseAttributes(BaseModel):
    hostname: Optional[str] = None
    port: Optional[int] = None
    username: Optional[str] = None
    password: Optional[str] = None
    platform: Optional[str] = None

    class Config:
        ignore_extra = False


class ConnectionOptions(BaseAttributes):
    extras: Optional[Dict[str, Any]]

    @classmethod
    def serialize(cls, i: inventory.ConnectionOptions) -> "ConnectionOptions":
        return ConnectionOptions(
            hostname=i.hostname,
            port=i.port,
            username=i.username,
            password=i.password,
            platform=i.platform,
            extras=i.extras,
        )


class InventoryElement(BaseAttributes):
    groups: List[str] = []
    data: Dict[str, Any] = {}
    connection_options: Dict[str, ConnectionOptions] = {}

    @classmethod
    def deserialize(
        cls,
        name: str,
        hostname: Optional[str] = None,
        port: Optional[int] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        platform: Optional[str] = None,
        groups: Optional[List[str]] = None,
        data: Optional[Dict[str, Any]] = None,
        connection_options: Optional[Dict[str, ConnectionOptions]] = None,
        defaults: inventory.Defaults = None,
    ) -> Dict[str, Any]:
        parent_groups = inventory.ParentGroups(groups)
        connection_options = connection_options or {}
        conn_opts = {
            k: inventory.ConnectionOptions(**v) for k, v in connection_options.items()
        }
        return {
            "name": name,
            "hostname": hostname,
            "port": port,
            "username": username,
            "password": password,
            "platform": platform,
            "groups": parent_groups,
            "data": data,
            "connection_options": conn_opts,
            "defaults": defaults,
        }

    @classmethod
    def deserialize_host(cls, **kwargs: Any) -> inventory.Host:
        return inventory.Host(**cls.deserialize(**kwargs))

    @classmethod
    def deserialize_group(cls, **kwargs: Any) -> inventory.Group:
        return inventory.Group(**cls.deserialize(**kwargs))

    @classmethod
    def serialize(cls, e: Union[inventory.Host, inventory.Group]) -> "InventoryElement":
        d = {}
        for f in cls.__fields__:
            d[f] = object.__getattribute__(e, f)
        d["groups"] = list(d["groups"])
        d["connection_options"] = {
            k: ConnectionOptions.serialize(v)
            for k, v in d["connection_options"].items()
        }
        return InventoryElement(**d)


class Defaults(BaseAttributes):
    data: Dict[str, Any] = {}
    connection_options: Dict[str, ConnectionOptions] = {}

    @classmethod
    def serialize(cls, defaults: inventory.Defaults) -> "InventoryElement":
        d = {}
        for f in cls.__fields__:
            d[f] = getattr(defaults, f)

        d["connection_options"] = {
            k: ConnectionOptions.serialize(v)
            for k, v in d["connection_options"].items()
        }
        return Defaults(**d)


class Inventory(BaseModel):
    hosts: Dict[str, InventoryElement]
    groups: Dict[str, InventoryElement]
    defaults: Defaults

    @classmethod
    def deserialize(
        cls,
        transform_function: Optional[Callable[..., Any]] = None,
        transform_function_options: Optional[Dict[str, Any]] = None,
        *args: Any,
        **kwargs: Any
    ) -> inventory.Inventory:
        transform_function_options = transform_function_options or {}
        deserialized = cls(*args, **kwargs)

        defaults_dict = deserialized.defaults.dict()
        for k, v in defaults_dict["connection_options"].items():
            defaults_dict["connection_options"][k] = inventory.ConnectionOptions(**v)
        defaults = inventory.Defaults(**defaults_dict)

        hosts = inventory.Hosts()
        for n, h in deserialized.hosts.items():
            hosts[n] = InventoryElement.deserialize_host(
                defaults=defaults, name=n, **h.dict()
            )

        groups = inventory.Groups()
        for n, g in deserialized.groups.items():
            groups[n] = InventoryElement.deserialize_group(name=n, **g.dict())

        return inventory.Inventory(
            hosts=hosts,
            groups=groups,
            defaults=defaults,
            transform_function=transform_function,
            transform_function_options=transform_function_options,
        )

    @classmethod
    def serialize(cls, inv: inventory.Inventory) -> "Inventory":
        hosts = {}
        for n, h in inv.hosts.items():
            hosts[n] = InventoryElement.serialize(h)
        groups = {}
        for n, g in inv.groups.items():
            groups[n] = InventoryElement.serialize(g)
        defaults = Defaults.serialize(inv.defaults)
        return Inventory(hosts=hosts, groups=groups, defaults=defaults)



import os
from typing import Any

from nornir.core.deserializer.inventory import Inventory, InventoryElement

import requests


class NSOTInventory(Inventory):
    """
    Inventory plugin that uses `nsot <https://github.com/dropbox/nsot>`_ as backend.

    Note:
        An extra attribute ``site`` will be assigned to the host. The value will be
        the name of the site the host belongs to.

    Environment Variables:
        * ``NSOT_URL``: Corresponds to nsot_url argument
        * ``NSOT_EMAIL``: Corresponds to nsot_email argument
        * ``NSOT_AUTH_HEADER``: Corresponds to nsot_auth_header argument
        * ``NSOT_SECRET_KEY``: Corresponds to nsot_secret_key argument

    Arguments:
        flatten_attributes: Assign host attributes to the root object. Useful
            for filtering hosts.
        nsot_url: URL to nsot's API (defaults to ``http://localhost:8990/api``)
        nsot_email: email for authtication (defaults to admin@acme.com)
        nsot_auth_header: String for auth_header authentication (defaults to X-NSoT-Email)
        nsot_secret_key: Secret Key for auth_token method. If given auth_token
            will be used as auth_method.
    """

    def __init__(
        self,
        nsot_url: str = "",
        nsot_email: str = "",
        nsot_secret_key: str = "",
        nsot_auth_header: str = "",
        flatten_attributes: bool = True,
        *args: Any,
        **kwargs: Any
    ) -> None:
        nsot_url = nsot_url or os.environ.get("NSOT_URL", "http://localhost:8990/api")
        nsot_email = nsot_email or os.environ.get("NSOT_EMAIL", "admin@acme.com")
        secret_key = nsot_secret_key or os.environ.get("NSOT_SECRET_KEY")

        if secret_key:
            data = {"email": nsot_email, "secret_key": secret_key}
            res = requests.post("{}/authenticate/".format(nsot_url), data=data)
            auth_token = res.json().get("auth_token")
            headers = {
                "Authorization": "AuthToken {}:{}".format(nsot_email, auth_token)
            }

        else:
            nsot_auth_header = nsot_auth_header or os.environ.get(
                "NSOT_AUTH_HEADER", "X-NSoT-Email"
            )
            headers = {nsot_auth_header: nsot_email}

        devices = requests.get("{}/devices".format(nsot_url), headers=headers).json()
        sites = requests.get("{}/sites".format(nsot_url), headers=headers).json()
        interfaces = requests.get(
            "{}/interfaces".format(nsot_url), headers=headers
        ).json()

        # We resolve site_id and assign "site" variable with the name of the site
        for d in devices:
            d["data"] = {"site": sites[d["site_id"] - 1]["name"], "interfaces": {}}

            remove_keys = []
            for k, v in d.items():
                if k not in InventoryElement().fields:
                    remove_keys.append(k)
                    d["data"][k] = v
            for r in remove_keys:
                d.pop(r)

            if flatten_attributes:
                # We assign attributes to the root
                for k, v in d["data"].pop("attributes").items():
                    d["data"][k] = v

        # We assign the interfaces to the hosts
        for i in interfaces:
            devices[i["device"] - 1]["data"]["interfaces"][i["name"]] = i

        # Finally the inventory expects a dict of hosts where the key is the hostname
        hosts = {d["hostname"]: d for d in devices}
        super().__init__(hosts=hosts, groups={}, defaults={}, *args, **kwargs)



import logging
import logging.handlers
import sys
import warnings
from pathlib import Path
from typing import Any, Callable, Dict, Optional, TYPE_CHECKING, Type, List

from nornir.core.exceptions import ConflictingConfigurationWarning

if TYPE_CHECKING:
    from nornir.core.deserializer.inventory import Inventory  # noqa


class SSHConfig(object):
    __slots__ = "config_file"

    def __init__(self, config_file: str) -> None:
        self.config_file = config_file


class InventoryConfig(object):
    __slots__ = "plugin", "options", "transform_function", "transform_function_options"

    def __init__(
        self,
        plugin: Type["Inventory"],
        options: Dict[str, Any],
        transform_function: Optional[Callable[..., Any]],
        transform_function_options: Optional[Dict[str, Any]],
    ) -> None:
        self.plugin = plugin
        self.options = options
        self.transform_function = transform_function
        self.transform_function_options = transform_function_options


class LoggingConfig(object):
    __slots__ = "enabled", "level", "file", "format", "to_console", "loggers"

    def __init__(
        self,
        enabled: Optional[bool],
        level: str,
        file_: str,
        format_: str,
        to_console: bool,
        loggers: List[str],
    ) -> None:
        self.enabled = enabled
        self.level = level
        self.file = file_
        self.format = format_
        self.to_console = to_console
        self.loggers = loggers

    def configure(self) -> None:
        if not self.enabled:
            return

        root_logger = logging.getLogger()
        if root_logger.hasHandlers() or root_logger.level != logging.WARNING:
            msg = (
                "Native Python logging configuration has been detected, but Nornir "
                "logging is enabled too. "
                "This can lead to unexpected logging results. "
                "Please set logging.enabled config to False "
                "to disable automatic Nornir logging configuration. Refer to "
                "https://nornir.readthedocs.io/en/stable/configuration/index.html#logging"  # noqa
            )
            warnings.warn(msg, ConflictingConfigurationWarning)

        formatter = logging.Formatter(self.format)
        # log INFO and DEBUG to stdout
        stdout_handler = logging.StreamHandler(sys.stdout)
        stdout_handler.setFormatter(formatter)
        stdout_handler.setLevel(logging.DEBUG)
        stdout_handler.addFilter(lambda record: record.levelno <= logging.INFO)
        # log WARNING, ERROR and CRITICAL to stderr
        stderr_handler = logging.StreamHandler(sys.stderr)
        stderr_handler.setFormatter(formatter)
        stderr_handler.setLevel(logging.WARNING)

        for logger_name in self.loggers:
            logger_ = logging.getLogger(logger_name)
            logger_.propagate = False
            logger_.setLevel(self.level)
            if logger_.hasHandlers():
                # Don't add handlers if some handlers are already attached to the logger
                # This is crucial to avoid duplicate handlers
                # Alternative would be to clear all handlers and reconfigure them
                # with Nornir
                # There are several situations this branch can be executed:
                # multiple calls to InitNornir,
                # logging.config.dictConfig configuring 'nornir' logger, etc.
                # The warning is not emitted in this scenario
                continue
            if self.file:
                handler = logging.handlers.RotatingFileHandler(
                    str(Path(self.file)), maxBytes=1024 * 1024 * 10, backupCount=20
                )
                handler.setFormatter(formatter)
                logger_.addHandler(handler)

            if self.to_console:
                logger_.addHandler(stdout_handler)
                logger_.addHandler(stderr_handler)


class Jinja2Config(object):
    __slots__ = "filters"

    def __init__(self, filters: Optional[Dict[str, Callable[..., Any]]]) -> None:
        self.filters = filters or {}


class CoreConfig(object):
    __slots__ = ("num_workers", "raise_on_error")

    def __init__(self, num_workers: int, raise_on_error: bool) -> None:
        self.num_workers = num_workers
        self.raise_on_error = raise_on_error


class Config(object):
    __slots__ = ("core", "ssh", "inventory", "jinja2", "logging", "user_defined")

    def __init__(
        self,
        inventory: InventoryConfig,
        ssh: SSHConfig,
        logging: LoggingConfig,
        jinja2: Jinja2Config,
        core: CoreConfig,
        user_defined: Dict[str, Any],
    ) -> None:
        self.inventory = inventory
        self.ssh = ssh
        self.logging = logging
        self.jinja2 = jinja2
        self.core = core
        self.user_defined = user_defined



import logging
import logging.config
from multiprocessing.dummy import Pool

from nornir.core.configuration import Config
from nornir.core.inventory import Inventory
from nornir.core.state import GlobalState
from nornir.core.task import AggregatedResult, Task

logger = logging.getLogger(__name__)


class Nornir(object):
    """
    This is the main object to work with. It contains the inventory and it serves
    as task dispatcher.

    Arguments:
        inventory (:obj:`nornir.core.inventory.Inventory`): Inventory to work with
        data(GlobalState): shared data amongst different iterations of nornir
        dry_run(``bool``): Whether if we are testing the changes or not
        config (:obj:`nornir.core.configuration.Config`): Configuration object

    Attributes:
        inventory (:obj:`nornir.core.inventory.Inventory`): Inventory to work with
        data(:obj:`nornir.core.GlobalState`): shared data amongst different iterations of nornir
        dry_run(``bool``): Whether if we are testing the changes or not
        config (:obj:`nornir.core.configuration.Config`): Configuration parameters
    """

    def __init__(
        self, inventory: Inventory, config: Config = None, data: GlobalState = None
    ) -> None:
        self.data = data if data is not None else GlobalState()

        self.inventory = inventory

        self.config = config or Config()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close_connections(on_good=True, on_failed=True)

    def filter(self, *args, **kwargs):
        """
        See :py:meth:`nornir.core.inventory.Inventory.filter`

        Returns:
            :obj:`Nornir`: A new object with same configuration as ``self`` but filtered inventory.
        """
        b = Nornir(**self.__dict__)
        b.inventory = self.inventory.filter(*args, **kwargs)
        return b

    def _run_serial(self, task, hosts, **kwargs):
        result = AggregatedResult(kwargs.get("name") or task.__name__)
        for host in hosts:
            result[host.name] = Task(task, **kwargs).start(host, self)
        return result

    def _run_parallel(self, task, hosts, num_workers, **kwargs):
        result = AggregatedResult(kwargs.get("name") or task.__name__)

        pool = Pool(processes=num_workers)
        result_pool = [
            pool.apply_async(Task(task, **kwargs).start, args=(h, self)) for h in hosts
        ]
        pool.close()
        pool.join()

        for rp in result_pool:
            r = rp.get()
            result[r.host.name] = r
        return result

    def run(
        self,
        task,
        num_workers=None,
        raise_on_error=None,
        on_good=True,
        on_failed=False,
        **kwargs,
    ):
        """
        Run task over all the hosts in the inventory.

        Arguments:
            task (``callable``): function or callable that will be run against each device in
              the inventory
            num_workers(``int``): Override for how many hosts to run in paralell for this task
            raise_on_error (``bool``): Override raise_on_error behavior
            on_good(``bool``): Whether to run or not this task on hosts marked as good
            on_failed(``bool``): Whether to run or not this task on hosts marked as failed
            **kwargs: additional argument to pass to ``task`` when calling it

        Raises:
            :obj:`nornir.core.exceptions.NornirExecutionError`: if at least a task fails
              and self.config.core.raise_on_error is set to ``True``

        Returns:
            :obj:`nornir.core.task.AggregatedResult`: results of each execution
        """
        num_workers = num_workers or self.config.core.num_workers

        run_on = []
        if on_good:
            for name, host in self.inventory.hosts.items():
                if name not in self.data.failed_hosts:
                    run_on.append(host)
        if on_failed:
            for name, host in self.inventory.hosts.items():
                if name in self.data.failed_hosts:
                    run_on.append(host)

        num_hosts = len(self.inventory.hosts)
        task_name = kwargs.get("name") or task.__name__
        if num_hosts:
            logger.info(
                f"Running task %r with args %s on %d hosts",
                task_name,
                kwargs,
                num_hosts,
            )
        else:
            logger.warning("Task %r has not been run – 0 hosts selected", task_name)

        if num_workers == 1:
            result = self._run_serial(task, run_on, **kwargs)
        else:
            result = self._run_parallel(task, run_on, num_workers, **kwargs)

        raise_on_error = (
            raise_on_error
            if raise_on_error is not None
            else self.config.core.raise_on_error
        )  # noqa
        if raise_on_error:
            result.raise_on_error()
        else:
            self.data.failed_hosts.update(result.failed_hosts.keys())
        return result

    def dict(self):
        """ Return a dictionary representing the object. """
        return {"data": self.data.dict(), "inventory": self.inventory.dict()}

    def close_connections(self, on_good=True, on_failed=False):
        def close_connections_task(task):
            task.host.close_connections()

        self.run(task=close_connections_task, on_good=on_good, on_failed=on_failed)

    @classmethod
    def get_validators(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):
        if not isinstance(v, cls):
            raise ValueError(f"Nornir: Nornir expected not {type(v)}")
        return v

    @property
    def state(self):
        return GlobalState



import importlib
import logging
from pathlib import Path
from typing import Any, Callable, Dict, Optional, Type, Union, List, cast

from nornir.core import configuration
from nornir.core.deserializer.inventory import Inventory

from pydantic import BaseSettings, Schema

import ruamel.yaml


logger = logging.getLogger(__name__)


class BaseNornirSettings(BaseSettings):
    def _build_values(self, init_kwargs: Dict[str, Any]) -> Dict[str, Any]:
        config_settings = init_kwargs.pop("__config_settings__", {})
        return {**config_settings, **self._build_environ(), **init_kwargs}


class SSHConfig(BaseNornirSettings):
    config_file: str = Schema(
        default="~/.ssh/config", description="Path to ssh configuration file"
    )

    class Config:
        env_prefix = "NORNIR_SSH_"
        ignore_extra = False

    @classmethod
    def deserialize(cls, **kwargs: Any) -> configuration.SSHConfig:
        s = SSHConfig(**kwargs)
        s.config_file = str(Path(s.config_file).expanduser())
        return configuration.SSHConfig(**s.dict())


class InventoryConfig(BaseNornirSettings):
    plugin: str = Schema(
        default="nornir.plugins.inventory.simple.SimpleInventory",
        description="Import path to inventory plugin",
    )
    options: Dict[str, Any] = Schema(
        default={}, description="kwargs to pass to the inventory plugin"
    )
    transform_function: str = Schema(
        default="",
        description=(
            "Path to transform function. The transform_function "
            "you provide will run against each host in the inventory"
        ),
    )
    transform_function_options: Dict[str, Any] = Schema(
        default={}, description="kwargs to pass to the transform_function"
    )

    class Config:
        env_prefix = "NORNIR_INVENTORY_"
        ignore_extra = False

    @classmethod
    def deserialize(cls, **kwargs: Any) -> configuration.InventoryConfig:
        inv = InventoryConfig(**kwargs)
        return configuration.InventoryConfig(
            plugin=cast(Type[Inventory], _resolve_import_from_string(inv.plugin)),
            options=inv.options,
            transform_function=_resolve_import_from_string(inv.transform_function),
            transform_function_options=inv.transform_function_options,
        )


class LoggingConfig(BaseNornirSettings):
    enabled: Optional[bool] = Schema(
        default=None, description="Whether to configure logging or not"
    )
    level: str = Schema(default="INFO", description="Logging level")
    file: str = Schema(default="nornir.log", description="Logging file")
    format: str = Schema(
        default="%(asctime)s - %(name)12s - %(levelname)8s - %(funcName)10s() - %(message)s",
        description="Logging format",
    )
    to_console: bool = Schema(
        default=False, description="Whether to log to console or not"
    )
    loggers: List[str] = Schema(default=["nornir"], description="Loggers to configure")

    class Config:
        env_prefix = "NORNIR_LOGGING_"
        ignore_extra = False

    @classmethod
    def deserialize(cls, **kwargs) -> configuration.LoggingConfig:
        conf = cls(**kwargs)
        return configuration.LoggingConfig(
            enabled=conf.enabled,
            level=conf.level.upper(),
            file_=conf.file,
            format_=conf.format,
            to_console=conf.to_console,
            loggers=conf.loggers,
        )


class Jinja2Config(BaseNornirSettings):
    filters: str = Schema(
        default="", description="Path to callable returning jinja filters to be used"
    )

    class Config:
        env_prefix = "NORNIR_JINJA2_"
        ignore_extra = False

    @classmethod
    def deserialize(cls, **kwargs: Any) -> configuration.Jinja2Config:
        c = Jinja2Config(**kwargs)
        jinja_filter_func = _resolve_import_from_string(c.filters)
        jinja_filters = jinja_filter_func() if jinja_filter_func else {}
        return configuration.Jinja2Config(filters=jinja_filters)


class CoreConfig(BaseNornirSettings):
    num_workers: int = Schema(
        default=20,
        description="Number of Nornir worker threads that are run at the same time by default",
    )
    raise_on_error: bool = Schema(
        default=False,
        description=(
            "If set to ``True``, (:obj:`nornir.core.Nornir.run`) method of "
            "will raise exception :obj:`nornir.core.exceptions.NornirExecutionError` "
            "if at least a host failed"
        ),
    )

    class Config:
        env_prefix = "NORNIR_CORE_"
        ignore_extra = False

    @classmethod
    def deserialize(cls, **kwargs: Any) -> configuration.CoreConfig:
        c = CoreConfig(**kwargs)
        return configuration.CoreConfig(**c.dict())


class Config(BaseNornirSettings):
    core: CoreConfig = CoreConfig()
    inventory: InventoryConfig = InventoryConfig()
    ssh: SSHConfig = SSHConfig()
    logging: LoggingConfig = LoggingConfig()
    jinja2: Jinja2Config = Jinja2Config()
    user_defined: Dict[str, Any] = Schema(
        default={}, description="User-defined <k, v> pairs"
    )

    class Config:
        env_prefix = "NORNIR_"
        ignore_extra = False

    @classmethod
    def deserialize(
        cls, __config_settings__: Optional[Dict[str, Any]] = None, **kwargs: Any
    ) -> configuration.Config:
        __config_settings__ = __config_settings__ or {}
        c = Config(
            core=CoreConfig(
                __config_settings__=__config_settings__.pop("core", {}),
                **kwargs.pop("core", {}),
            ),
            ssh=SSHConfig(
                __config_settings__=__config_settings__.pop("ssh", {}),
                **kwargs.pop("ssh", {}),
            ),
            inventory=InventoryConfig(
                __config_settings__=__config_settings__.pop("inventory", {}),
                **kwargs.pop("inventory", {}),
            ),
            logging=LoggingConfig(
                __config_settings__=__config_settings__.pop("logging", {}),
                **kwargs.pop("logging", {}),
            ),
            jinja2=Jinja2Config(
                __config_settings__=__config_settings__.pop("jinja2", {}),
                **kwargs.pop("jinja2", {}),
            ),
            __config_settings__=__config_settings__,
            **kwargs,
        )
        return configuration.Config(
            core=CoreConfig.deserialize(**c.core.dict()),
            inventory=InventoryConfig.deserialize(**c.inventory.dict()),
            ssh=SSHConfig.deserialize(**c.ssh.dict()),
            logging=LoggingConfig.deserialize(**c.logging.dict()),
            jinja2=Jinja2Config.deserialize(**c.jinja2.dict()),
            user_defined=c.user_defined,
        )

    @classmethod
    def load_from_file(cls, config_file: str, **kwargs: Any) -> configuration.Config:
        config_dict: Dict[str, Any] = {}
        if config_file:
            yml = ruamel.yaml.YAML(typ="safe")
            with open(config_file, "r") as f:
                config_dict = yml.load(f) or {}
        return Config.deserialize(__config_settings__=config_dict, **kwargs)


def _resolve_import_from_string(
    import_path: Union[Callable[..., Any], str]
) -> Optional[Callable[..., Any]]:
    try:
        if not import_path:
            return None
        elif callable(import_path):
            return import_path
        module_name, obj_name = import_path.rsplit(".", 1)
        module = importlib.import_module(module_name)
        return getattr(module, obj_name)
    except Exception:
        logger.error("Failed to import %r", import_path, exc_info=True)
        raise



from typing import List

from nornir.core.task import Result, Task


def napalm_cli(task: Task, commands: List[str]) -> Result:
    """
    Run commands on remote devices using napalm

    Arguments:
        commands: commands to execute

    Returns:
        Result object with the following attributes set:
          * result (``dict``): result of the commands execution
    """
    device = task.host.get_connection("napalm", task.nornir.config)
    result = device.cli(commands)
    return Result(host=task.host, result=result)



from __future__ import unicode_literals

from nornir.core.task import Result, Task


def netmiko_save_config(
    task: Task, cmd: str = "", confirm: bool = False, confirm_response: str = ""
) -> Result:
    """
    Execute Netmiko save_config method
    Arguments:
        cmd(str, optional): Command used to save the configuration.
        confirm(bool, optional): Does device prompt for confirmation before executing save operation
        confirm_response(str, optional): Response send to device when it prompts for confirmation

    Returns:
        :obj: `nornir.core.task.Result`:
          * result (``str``): String showing the CLI output from the save operation
    """
    conn = task.host.get_connection("netmiko", task.nornir.config)
    if cmd:
        result = conn.save_config(
            cmd=cmd, confirm=confirm, confirm_response=confirm_response
        )
    else:
        result = conn.save_config(confirm=confirm, confirm_response=confirm_response)
    return Result(host=task.host, result=result, changed=True)



from typing import Any, Optional, Dict, Callable

from nornir.core.helpers import jinja_helper
from nornir.core.task import Result, Task

FiltersDict = Optional[Dict[str, Callable[..., str]]]


def template_string(
    task: Task, template: str, jinja_filters: FiltersDict = None, **kwargs: Any
) -> Result:
    """
    Renders a string with jinja2. All the host data is available in the template

    Arguments:
        template (string): template string
        jinja_filters (dict): jinja filters to enable. Defaults to nornir.config.jinja2.filters
        **kwargs: additional data to pass to the template

    Returns:
        Result object with the following attributes set:
          * result (``string``): rendered string
    """
    jinja_filters = jinja_filters or {} or task.nornir.config.jinja2.filters
    text = jinja_helper.render_from_string(
        template=template, host=task.host, jinja_filters=jinja_filters, **kwargs
    )
    return Result(host=task.host, result=text)



import os
from typing import Any, Dict, Optional

from nornir.core.deserializer.inventory import Inventory, HostsDict


import requests


class NBInventory(Inventory):
    def __init__(
        self,
        nb_url: Optional[str] = None,
        nb_token: Optional[str] = None,
        use_slugs: bool = True,
        flatten_custom_fields: bool = True,
        filter_parameters: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> None:
        """
        Netbox plugin

        Arguments:
            nb_url: Netbox url, defaults to http://localhost:8080.
                You can also use env variable NB_URL
            nb_token: Netbokx token. You can also use env variable NB_TOKEN
            use_slugs: Whether to use slugs or not
            flatten_custom_fields: Whether to assign custom fields directly to the host or not
            filter_parameters: Key-value pairs to filter down hosts
        """
        filter_parameters = filter_parameters or {}

        nb_url = nb_url or os.environ.get("NB_URL", "http://localhost:8080")
        nb_token = nb_token or os.environ.get(
            "NB_TOKEN", "0123456789abcdef0123456789abcdef01234567"
        )
        headers = {"Authorization": "Token {}".format(nb_token)}

        # Create dict of hosts using 'devices' from NetBox
        r = requests.get(
            "{}/api/dcim/devices/?limit=0".format(nb_url),
            headers=headers,
            params=filter_parameters,
        )
        r.raise_for_status()
        nb_devices = r.json()

        hosts = {}
        for d in nb_devices["results"]:
            host: HostsDict = {"data": {}}

            # Add value for IP address
            if d.get("primary_ip", {}):
                host["hostname"] = d["primary_ip"]["address"].split("/")[0]

            # Add values that don't have an option for 'slug'
            host["data"]["serial"] = d["serial"]
            host["data"]["vendor"] = d["device_type"]["manufacturer"]["name"]
            host["data"]["asset_tag"] = d["asset_tag"]

            if flatten_custom_fields:
                for cf, value in d["custom_fields"].items():
                    host["data"][cf] = value
            else:
                host["data"]["custom_fields"] = d["custom_fields"]

            # Add values that do have an option for 'slug'
            if use_slugs:
                host["data"]["site"] = d["site"]["slug"]
                host["data"]["role"] = d["device_role"]["slug"]
                host["data"]["model"] = d["device_type"]["slug"]

                # Attempt to add 'platform' based of value in 'slug'
                host["platform"] = d["platform"]["slug"] if d["platform"] else None

            else:
                host["data"]["site"] = d["site"]["name"]
                host["data"]["role"] = d["device_role"]
                host["data"]["model"] = d["device_type"]
                host["platform"] = d["platform"]

            # Assign temporary dict to outer dict
            hosts[d["name"]] = host

        # Pass the data back to the parent class
        super().__init__(hosts=hosts, groups={}, defaults={}, **kwargs)



from nornir.core.exceptions import CommandError
from nornir.core.task import Result, Task

from paramiko.agent import AgentRequestHandler


def remote_command(task: Task, command: str) -> Result:
    """
    Executes a command remotely on the host

    Arguments:
        command (``str``): command to execute

    Returns:
        Result object with the following attributes set:
          * result (``str``): stderr or stdout
          * stdout (``str``): stdout
          * stderr (``str``): stderr

    Raises:
        :obj:`nornir.core.exceptions.CommandError`: when there is a command error
    """
    client = task.host.get_connection("paramiko", task.nornir.config)
    connection_state = task.host.get_connection_state("paramiko")

    chan = client.get_transport().open_session()

    if connection_state["ssh_forward_agent"]:
        AgentRequestHandler(chan)

    chan.exec_command(command)

    with chan.makefile() as f:
        stdout = f.read().decode()
    with chan.makefile_stderr() as f:
        stderr = f.read().decode()

    exit_status_code = chan.recv_exit_status()

    if exit_status_code:
        raise CommandError(command, exit_status_code, stdout, stderr)

    result = stderr if stderr else stdout
    return Result(result=result, host=task.host, stderr=stderr, stdout=stdout)



from .napalm_cli import napalm_cli
from .napalm_configure import napalm_configure
from .napalm_get import napalm_get
from .napalm_validate import napalm_validate
from .netmiko_file_transfer import netmiko_file_transfer
from .netmiko_send_command import netmiko_send_command
from .netmiko_send_config import netmiko_send_config
from .netmiko_save_config import netmiko_save_config
from .tcp_ping import tcp_ping

__all__ = (
    "napalm_cli",
    "napalm_configure",
    "napalm_get",
    "napalm_validate",
    "netmiko_file_transfer",
    "netmiko_send_command",
    "netmiko_send_config",
    "netmiko_save_config",
    "tcp_ping",
)



import base64
import difflib
import threading
from pathlib import Path
from typing import Tuple

from nornir.core.task import Result, Task

import requests


LOCK = threading.Lock()


def _generate_diff(original: str, fromfile: str, tofile: str, content: str) -> str:
    diff = difflib.unified_diff(
        original.splitlines(), content.splitlines(), fromfile=fromfile, tofile=tofile
    )
    return "\n".join(diff)


def _get_repository(session: requests.Session, url: str, repository: str) -> int:
    resp = session.get(f"{url}/api/v4/projects?search={repository}")
    if resp.status_code != 200:
        raise RuntimeError(f"Unexpected Gitlab status code {resp.status_code}")

    pid = 0
    found = False
    respjson = resp.json()
    if not len(respjson):
        raise RuntimeError("Gitlab repository not found")

    for p in respjson:
        if p.get("name", "") == repository:
            found = True
            pid = p.get("id", 0)

    if not pid or not found:
        raise RuntimeError("Gitlab repository not found")

    return pid


def _remote_exists(
    task: Task, session: requests.Session, url: str, pid: int, filename: str, ref: str
) -> Tuple[bool, str]:
    resp = session.get(
        f"{url}/api/v4/projects/{pid}/repository/files/{filename}?ref={ref}"
    )
    if resp.status_code == 200:
        return (
            True,
            base64.decodebytes(resp.json()["content"].encode("ascii")).decode(),
        )
    return (False, "")


def _local_exists(task: Task, filename: str) -> Tuple[bool, str]:
    try:
        with open(Path(filename)) as f:
            content = f.read()
        return (True, content)
    except FileNotFoundError:
        return (False, "")


def _create(
    task: Task,
    session: requests.Session,
    url: str,
    pid: int,
    filename: str,
    content: str,
    branch: str,
    commit_message: str,
    dry_run: bool,
) -> str:
    if dry_run:
        return _generate_diff("", "", filename, content)

    with LOCK:
        url = f"{url}/api/v4/projects/{pid}/repository/files/{filename}"
        data = {"branch": branch, "content": content, "commit_message": commit_message}
        resp = session.post(url, data=data)

        if resp.status_code != 201:
            raise RuntimeError(f"Unable to create file: {filename}!")
    return _generate_diff("", "", filename, content)


def _update(
    task: Task,
    session: requests.Session,
    url: str,
    pid: int,
    filename: str,
    content: str,
    branch: str,
    commit_message: str,
    dry_run: bool,
) -> str:
    exists, original = _remote_exists(task, session, url, pid, filename, branch)

    if not exists:
        raise RuntimeError(f"File '{filename}' does not exist!")

    if dry_run:
        return _generate_diff(original, filename, filename, content)

    if original != content:
        with LOCK:
            url = f"{url}/api/v4/projects/{pid}/repository/files/{filename}"
            data = {
                "branch": branch,
                "content": content,
                "commit_message": commit_message,
            }
            resp = session.put(url=url, data=data)
            if resp.status_code != 200:
                raise RuntimeError(f"Unable to update file: {filename}")
    return _generate_diff(original, filename, filename, content)


def _get(
    task: Task,
    session: requests.Session,
    url: str,
    pid: int,
    filename: str,
    destination: str,
    ref: str,
    dry_run: bool,
) -> str:

    # if destination is not provided, use the filename as destination in current
    # directory
    if destination == "":
        destination = filename

    (_, local) = _local_exists(task, destination)

    (status, content) = _remote_exists(task, session, url, pid, filename, ref)

    if not status:
        raise RuntimeError(f"Unable to get file: {filename}")

    if not dry_run:
        if local != content:
            with open(destination, "w") as f:
                f.write(content)

    return _generate_diff(local, destination, destination, content)


def gitlab(
    task: Task,
    url: str,
    token: str,
    repository: str,
    filename: str,
    content: str = "",
    action: str = "create",
    dry_run: bool = False,
    branch: str = "master",
    destination: str = "",
    ref: str = "master",
    commit_message: str = "",
) -> Result:
    """
    Exposes some of the Gitlab API functionality for operations on files
    in a Gitlab repository.

    Example:

        nornir.run(files.gitlab,
                   action="create",
                   url="https://gitlab.localhost.com",
                   token="ABCD1234",
                   repository="test",
                   filename="config",
                   ref="master")

    Arguments:
        dry_run: Whether to apply changes or not
        url: Gitlab instance URL
        token: Personal access token
        repository: source/destination repository
        filename: source/destination file name
        content: content to write
        action: ``create``, ``update``, ``get``
        branch: destination branch
        destination: local destination filename (only used in get action)
        ref: branch, commit hash or tag (only used in get action)
        commit_message: commit message

    Returns:
        Result object with the following attributes set:
            * changed (``bool``):
            * diff (``str``): unified diff

    """
    dry_run = task.is_dry_run(dry_run)

    session = requests.session()
    session.headers.update({"PRIVATE-TOKEN": token})

    if commit_message == "":
        commit_message = "File created with nornir"

    pid = _get_repository(session, url, repository)

    if action == "create":
        diff = _create(
            task, session, url, pid, filename, content, branch, commit_message, dry_run
        )
    elif action == "update":
        diff = _update(
            task, session, url, pid, filename, content, branch, commit_message, dry_run
        )
    elif action == "get":
        diff = _get(task, session, url, pid, filename, destination, ref, dry_run)
    return Result(host=task.host, diff=diff, changed=bool(diff))



import hashlib
import os
import stat
from typing import List, Optional

from nornir.core.exceptions import CommandError
from nornir.core.task import Result, Task
from nornir.plugins.tasks import commands

import paramiko

from scp import SCPClient


def get_src_hash(filename: str) -> str:
    sha1sum = hashlib.sha1()

    with open(filename, "rb") as f:
        block = f.read(2 ** 16)
        while len(block) != 0:
            sha1sum.update(block)
            block = f.read(2 ** 16)
    return sha1sum.hexdigest()


def get_dst_hash(task: Task, filename: str) -> str:
    command = "sha1sum {}".format(filename)
    try:
        result = commands.remote_command(task, command)
        if result.stdout is not None:
            return result.stdout.split()[0]

    except CommandError as e:
        if "No such file or directory" in e.stderr:
            return ""

        raise

    return ""


def remote_exists(sftp_client: paramiko.SFTPClient, f: str) -> bool:
    try:
        sftp_client.stat(f)
        return True

    except IOError:
        return False


def compare_put_files(
    task: Task, sftp_client: paramiko.SFTPClient, src: str, dst: str
) -> List[str]:
    changed = []
    if os.path.isfile(src):
        src_hash = get_src_hash(src)
        try:
            dst_hash = get_dst_hash(task, dst)
        except IOError:
            dst_hash = ""
        if src_hash != dst_hash:
            changed.append(dst)
    else:
        if remote_exists(sftp_client, dst):
            for f in os.listdir(src):
                s = os.path.join(src, f)
                d = os.path.join(dst, f)
                changed.extend(compare_put_files(task, sftp_client, s, d))
        else:
            changed.append(dst)
    return changed


def compare_get_files(
    task: Task, sftp_client: paramiko.SFTPClient, src: str, dst: str
) -> List[str]:
    changed = []
    if stat.S_ISREG(sftp_client.stat(src).st_mode):
        # is a file
        src_hash = get_dst_hash(task, src)
        try:
            dst_hash = get_src_hash(dst)
        except IOError:
            dst_hash = ""
        if src_hash != dst_hash:
            changed.append(dst)
    else:
        if os.path.exists(dst):
            for f in sftp_client.listdir(src):
                s = os.path.join(src, f)
                d = os.path.join(dst, f)
                changed.extend(compare_get_files(task, sftp_client, s, d))
        else:
            changed.append(dst)
    return changed


def get(
    task: Task,
    scp_client: SCPClient,
    sftp_client: paramiko.SFTPClient,
    src: str,
    dst: str,
    dry_run: Optional[bool] = None,
) -> List[str]:
    changed = compare_get_files(task, sftp_client, src, dst)
    if changed and not dry_run:
        scp_client.get(src, dst, recursive=True)
    return changed


def put(
    task: Task,
    scp_client: SCPClient,
    sftp_client: paramiko.SFTPClient,
    src: str,
    dst: str,
    dry_run: Optional[bool] = None,
) -> List[str]:
    changed = compare_put_files(task, sftp_client, src, dst)
    if changed and not dry_run:
        scp_client.put(src, dst, recursive=True)
    return changed


def sftp(
    task: Task, src: str, dst: str, action: str, dry_run: Optional[bool] = None
) -> Result:
    """
    Transfer files from/to the device using sftp protocol

    Example::

        nornir.run(files.sftp,
                    action="put",
                    src="README.md",
                    dst="/tmp/README.md")

    Arguments:
        dry_run: Whether to apply changes or not
        src: source file
        dst: destination
        action: ``put``, ``get``.

    Returns:
        Result object with the following attributes set:
          * changed (``bool``):
          * files_changed (``list``): list of files that changed
    """
    dry_run = task.is_dry_run(dry_run)
    actions = {"put": put, "get": get}
    client = task.host.get_connection("paramiko", task.nornir.config)
    scp_client = SCPClient(client.get_transport())
    sftp_client = paramiko.SFTPClient.from_transport(client.get_transport())
    files_changed = actions[action](task, scp_client, sftp_client, src, dst, dry_run)
    return Result(
        host=task.host, changed=bool(files_changed), files_changed=files_changed
    )



from typing import Any, List, Optional

from nornir.core.task import Result, Task


def netmiko_send_config(
    task: Task,
    config_commands: Optional[List[str]] = None,
    config_file: Optional[str] = None,
    **kwargs: Any
) -> Result:
    """
    Execute Netmiko send_config_set method (or send_config_from_file)

    Arguments:
        config_commands: Commands to configure on the remote network device.
        config_file: File to read configuration commands from.
        kwargs: Additional arguments to pass to method.

    Returns:
        Result object with the following attributes set:
          * result (``str``): string showing the CLI from the configuration changes.
    """
    net_connect = task.host.get_connection("netmiko", task.nornir.config)
    net_connect.enable()
    if config_commands:
        result = net_connect.send_config_set(config_commands=config_commands, **kwargs)
    elif config_file:
        result = net_connect.send_config_from_file(config_file=config_file, **kwargs)
    else:
        raise ValueError("Must specify either config_commands or config_file")

    return Result(host=task.host, result=result, changed=True)



import difflib
import os
from typing import List, Optional

from nornir.core.task import Result, Task


def _read_file(file: str) -> List[str]:
    if not os.path.exists(file):
        return []

    with open(file, "r") as f:
        return f.read().splitlines()


def _generate_diff(filename: str, content: str, append: bool) -> str:
    original = _read_file(filename)
    if append:
        c = list(original)
        c.extend(content.splitlines())
        new_content = c
    else:
        new_content = content.splitlines()

    diff = difflib.unified_diff(original, new_content, fromfile=filename, tofile="new")

    return "\n".join(diff)


def write_file(
    task: Task,
    filename: str,
    content: str,
    append: bool = False,
    dry_run: Optional[bool] = None,
) -> Result:
    """
    Write contents to a file (locally)

    Arguments:
        dry_run: Whether to apply changes or not
        filename: file you want to write into
        content: content you want to write
        append: whether you want to replace the contents or append to it

    Returns:
        Result object with the following attributes set:
          * changed (``bool``):
          * diff (``str``): unified diff
    """
    diff = _generate_diff(filename, content, append)

    if not task.is_dry_run(dry_run):
        mode = "a+" if append else "w+"
        with open(filename, mode=mode) as f:
            f.write(content)

    return Result(host=task.host, diff=diff, changed=bool(diff))
